unit GifLZW;

// Example of GIF-LZW compression/decompression plug-in for ImageEn

// The compressor is based on:
//   * ppmtogif.c (pbmplus) by Jef Poskanzer and others.
//   * gifcompr.c, gifencode.c (GIFENCOD) by David Rowley <mgardi@watdscu.waterloo.edu>.
//   * writegif.c (GIFTOOL) by David Koblas <koblas@netcom.com>
//   * compress.c - File compression ala IEEE Computer, June 1984, by
//     Spencer W. Thomas       (decvax!harpo!utah-cs!utah-gr!thomas)
//     Jim McKie               (decvax!mcvax!jim)
//     Steve Davies            (decvax!vax135!petsd!peora!srd)
//     Ken Turkowski           (decvax!decwrl!turtlevax!ken)
//     James A. Woods          (decvax!ihnp4!ames!jaw)
//     Joe Orost               (decvax!vax135!petsd!joe)
//   * gcd.c (gd) by Tom Boutell <boutell@boutell.com>
//
// The decompressor is based on work done by
//   * readgif.c (GIFTOOL) by David Koblas <koblas@netcom.com>

{$I IE.INC}

interface

uses Windows, Graphics, classes, sysutils;

procedure GIFLZWCompress(Stream: TStream; Height,Width:integer; Interlaced:boolean; FData:pchar; BitsPerPixel:integer);
procedure GIFLZWDecompress(Stream: TStream; Height,Width:integer; Interlaced:boolean; FData:pchar);

implementation

(*******************************************************************************
**
**			THashTable
**
*******************************************************************************)
const
  HashKeyBits		= 13;
  GIFCodeBits		= 12;
  HashSize		= 8009;
  StackSize		= (2 SHL GIFCodeBits);
  TableSize		= (1 SHL GIFCodeBits);
  HashKeyMax		= (1 SHL HashKeyBits)-1;
  GIFCodeMax		= (1 SHL GIFCodeBits)-1;
  TableMaxMaxCode	= (1 SHL GIFCodeBits);
  TableMaxFill		= TableMaxMaxCode-1;
  HashKeyMask		= HashKeyMax;
  GIFCodeMask		= GIFCodeMax;
  HashEmpty		= $000FFFFF;

type
  KeyInt = longInt;
  CodeInt = SmallInt;
  THashArray = array[0..HashSize-1] of KeyInt;
  PHashArray = ^THashArray;
  THashTable = class
	  HashTable: PHashArray;
	  public
		   constructor Create;
   	   destructor Destroy; override;
		   procedure Clear;
		   procedure Insert(Key: KeyInt; Code: CodeInt);
		   function Lookup(Key: KeyInt): CodeInt;
	  end;

/////////////////////////////////////////////////////////////////////////////////////
function HashKey(Key: KeyInt): CodeInt;
begin
  Result := ((Key SHR (GIFCodeBits-8)) XOR Key) MOD HashSize;
end;

/////////////////////////////////////////////////////////////////////////////////////
function NextHashKey(HKey: CodeInt): CodeInt;
var
  disp		: CodeInt;
begin
  disp := HashSize - HKey;
  if (HKey = 0) then
    disp := 1;
  dec(HKey, disp);
  if (HKey < 0) then
      inc(HKey, HashSize);
  Result := HKey;
end;

/////////////////////////////////////////////////////////////////////////////////////
constructor THashTable.Create;
begin
  inherited Create;
  GetMem(HashTable, sizeof(THashArray));
  Clear;
end;

/////////////////////////////////////////////////////////////////////////////////////
destructor THashTable.Destroy;
begin
  FreeMem(HashTable);
  inherited Destroy;
end;

/////////////////////////////////////////////////////////////////////////////////////
procedure THashTable.Clear;
begin
  FillChar(HashTable^, sizeof(THashArray), $FF);
end;

/////////////////////////////////////////////////////////////////////////////////////
procedure THashTable.Insert(Key: KeyInt; Code: CodeInt);
var
  HKey			: CodeInt;
begin
  HKey := HashKey(Key);
  while (HashTable[HKey] AND (HashEmpty SHL GIFCodeBits) <> (HashEmpty SHL GIFCodeBits)) do
    HKey := NextHashKey(HKey);
  HashTable[HKey] := (Key SHL GIFCodeBits) OR (Code AND GIFCodeMask);
end;

/////////////////////////////////////////////////////////////////////////////////////
function THashTable.Lookup(Key: KeyInt): CodeInt;
var
  HKey			: CodeInt;
  HTKey			: KeyInt;
begin
  HKey := HashKey(Key);
  Key := Key SHL GIFCodeBits;
  HTKey := HashTable[HKey] AND (HashEmpty SHL GIFCodeBits);
  while (HTKey <> HashEmpty SHL GIFCodeBits) do
  begin
    if (Key = HTKey) then
    begin
      Result := HashTable[HKey] AND GIFCodeMask;
      exit;
    end;
    HKey := NextHashKey(HKey);
    HTKey := HashTable[HKey] AND (HashEmpty SHL GIFCodeBits);
  end;
  Result := -1;
end;

/////////////////////////////////////////////////////////////////////////////////////
procedure ReadCheck(Stream: TStream; var Buffer; Size: LongInt);
begin
  if (Stream.Read(Buffer, Size) <> Size) then
   	raise EInvalidGraphic.create('unrecognized GIF');
end;

/////////////////////////////////////////////////////////////////////////////////////
procedure GIFLZWDecompress(Stream: TStream; Height,Width:integer; Interlaced:boolean; FData:pchar);
var
  table			: array[0..1, 0..TableSize-1] of integer;
  firstcode, oldcode	: integer;
  buf			: array[0..257] of BYTE;
  Dest			: PChar;
  v,xpos, ypos, pass	: integer;
  stack			: array[0..StackSize-1] of integer;
  Source			: ^integer;
  BitsPerCode		: integer;
  InitialBitsPerCode	: BYTE;
  MaxCode		: CodeInt;
  MaxCodeSize		: integer;
  ClearCode		: integer;
  EOFCode		: integer;
  step			: integer;
  i			: integer;
  StartBit,LastBit,LastByte		: integer;
  get_done,return_clear,ZeroBlock		: boolean;
  /////////////////////////////////////////////////
  function nextCode(BitsPerCode: integer): integer;
  const
    masks: array[0..15] of integer =
      ($0000, $0001, $0003, $0007,
       $000f, $001f, $003f, $007f,
       $00ff, $01ff, $03ff, $07ff,
       $0fff, $1fff, $3fff, $7fff);
  var
    StartIndex, EndIndex		: integer;
    ret			: integer;
    EndBit		: integer;
    count		: BYTE;
  begin
    if (return_clear) then
    begin
      return_clear := FALSE;
      Result := ClearCode;
      exit;
    end;
    EndBit := StartBit + BitsPerCode;
    if (EndBit >= LastBit) then
    begin
      if (get_done) then
      begin
        Result := -1;
        exit;
      end;
      buf[0] := buf[LastByte-2];
      buf[1] := buf[LastByte-1];
      if (Stream.Read(count, 1) <> 1) then
      begin
        Result := -1;
        exit;
      end;
      if (count = 0) then
      begin
        ZeroBlock := True;
        get_done := TRUE;
      end else
        ReadCheck(Stream, Buf[2], Count);
      LastByte := 2 + count;
      StartBit := (StartBit - LastBit) + 16;
      LastBit := LastByte * 8;
      EndBit := StartBit + BitsPerCode;
    end;
    EndIndex := EndBit DIV 8;
    StartIndex := StartBit DIV 8;
    if (StartIndex = EndIndex) then
      ret := buf[StartIndex]
    else
      if (StartIndex + 1 = EndIndex) then
        ret := buf[StartIndex] OR (buf[StartIndex+1] SHL 8)
      else
        ret := buf[StartIndex] OR (buf[StartIndex+1] SHL 8) OR (buf[StartIndex+2] SHL 16);
    ret := (ret SHR (StartBit AND $0007)) AND masks[BitsPerCode];
    Inc(StartBit, BitsPerCode);
    Result := ret;
  end;
  /////////////////////////////////////////////////
  function nextLWZ: integer;
  var
    code, incode	: integer;
    i			: integer;
    b			: BYTE;
  begin
    code := nextCode(BitsPerCode);
    while (code >= 0) do
    begin
      if (code = ClearCode) then
      begin
        for i := 0 to ClearCode-1 do
        begin
          table[0, i] := 0;
          table[1, i] := i;
        end;
        for i := ClearCode to TableSize-1 do
        begin
          table[0, i] := 0;
          table[1, i] := 0;
        end;
        BitsPerCode := InitialBitsPerCode+1;
        MaxCodeSize := 2 * ClearCode;
        MaxCode := ClearCode + 2;
        Source := @stack;
        repeat
          firstcode := nextCode(BitsPerCode);
          oldcode := firstcode;
        until (firstcode <> ClearCode);
        Result := firstcode;
        exit;
      end;
      if (code = EOFCode) then
      begin
        Result := -2;
        if (ZeroBlock) then
          exit;
        if (Stream.Read(b, 1) <> 1) then
          exit;
        while (b <> 0) do
        begin
          Stream.Seek(b, soFromCurrent);
          if (Stream.Read(b, 1) <> 1) then
            exit;
        end;
        exit;
      end;
      incode := code;
      if (code >= MaxCode) then
      begin
        Source^ := firstcode;
        Inc(Source);
        code := oldcode;
      end;
      while (code >= ClearCode) do
      begin
        Source^ := table[1, code];
        Inc(Source);
        if (code = table[0, code]) then
        	 raise EInvalidGraphic.create('unrecognized GIF');
        code := table[0, code];
      end;
      firstcode := table[1, code];
      Source^ := firstcode;
      Inc(Source);
      code := MaxCode;
      if (code <= GIFCodeMax) then
      begin
        table[0, code] := oldcode;
        table[1, code] := firstcode;
        Inc(MaxCode);
        if ((MaxCode >= MaxCodeSize) and (MaxCodeSize <= GIFCodeMax)) then
        begin
          MaxCodeSize := MaxCodeSize * 2;
          Inc(BitsPerCode);
        end;
      end;
      oldcode := incode;
      if (longInt(Source) > longInt(@stack)) then
      begin
        Dec(Source);
        Result := Source^;
        exit;
      end
    end;
    Result := code;
  end;
  /////////////////////////////////////////////////
  function readLWZ: integer;
  begin
    if (longInt(Source) > longInt(@stack)) then
    begin
      Dec(Source);
      Result := Source^;
    end else
      Result := nextLWZ;
  end;
/////////////////////////////////////////////////
begin
  if (Stream.Read(InitialBitsPerCode, 1) <> 1) then
    exit;
  BitsPerCode := InitialBitsPerCode + 1;
  ClearCode := 1 SHL InitialBitsPerCode;
  EOFCode := ClearCode + 1;
  MaxCodeSize := 2 * ClearCode;
  MaxCode := ClearCode + 2;
  StartBit := 0;
  LastBit := 0;
  LastByte := 2;
  ZeroBlock := False;
  get_done := FALSE;
  return_clear := TRUE;
  Source := @stack;
  try
    if (Interlaced) then
    begin
      ypos := 0;
      pass := 0;
      step := 8;
      for i := 0 to Height-1 do
      begin
        Dest := FData + Width * ypos;
        for xpos := 0 to width-1 do
        begin
          v := readLWZ;
          if (v < 0) then
            exit;
          Dest^ := char(v);
          Inc(Dest);
        end;
        Inc(ypos, step);
        if (ypos >= height) then
          repeat
            if (pass > 0) then
              step := step DIV 2;
            Inc(pass);
            ypos := step DIV 2;
          until (ypos <= height);
      end;
    end else
    begin
      Dest := FData;
      for ypos := 0 to (height * width)-1 do
      begin
        v := readLWZ;
        if (v < 0) then
          exit;
        Dest^ := char(v);
        Inc(Dest);
      end;
    end;
  finally
  end;
end;

/////////////////////////////////////////////////////////////////////////////////////
procedure WriteByte(Stream: TStream; b: BYTE);
begin
  Stream.Write(b, 1);
end;

/////////////////////////////////////////////////////////////////////////////////////
procedure GIFLZWCompress(Stream: TStream; Height,Width:integer; Interlaced:boolean; FData:pchar; BitsPerPixel:integer);
const
  EndBlockByte		= $00;
var
  bpp			: integer;
  Pixel			: PChar;
  PixelVal		: char;
  Prefix		: CodeInt;
  cX			: LongInt;
  Y			: LongInt;
  Pass			: integer;
  HashTable		: THashTable;
  BitsPerCode		: integer;
  MaxCode		: CodeInt;
  NewKey		: KeyInt;
  NewCode		: CodeInt;
  FreeEntry		: CodeInt;
  OutputBucket		: longInt;
  OutputBits		: integer;
  ClearFlag		: Boolean;
  InitialBitsPerCode	: integer;
  ClearCode		: integer;
  EOFCode		: integer;
  Buffer		: TMemoryStream;
  fdatasize:integer;
  /////////////////////////////
  function BumpPixel: boolean;
  begin
    dec(cX);
    if (cX <= 0) then
    begin
      if not(Interlaced) then
      begin
        Result := False;
        exit;
      end;
      cX := Width;
      case (Pass) of
        0:
          begin
            inc(Y, 8);
            if (Y >= Height) then
            begin
              inc(Pass);
              Y := 4;
            end;
          end;
        1:
          begin
            inc(Y, 8);
            if (Y >= Height) then
            begin
              inc(Pass);
              Y := 2;
            end;
          end;
        2:
          begin
            inc(Y, 4);
            if (Y >= Height) then
            begin
              inc(Pass);
              Y := 1;
            end;
          end;
        3:
          inc(Y, 2);
      end;
      if (Y >= height) then
      begin
        Result := False;
        exit;
      end;
      Pixel := FData + (Y * Width);
    end;
    Result := True;
  end;
  //////////////////////////////////////////////////
  function MaxCodesFromBits(Bits: integer): CodeInt;
  begin
    Result := (CodeInt(1) SHL Bits) - 1;
  end;
  /////////////////////////////////
  procedure output(code: CodeInt);
  const
    BitBucketMask: array[0..16] of longInt =
      ($0000,
       $0001, $0003, $0007, $000F,
       $001F, $003F, $007F, $00FF,
       $01FF, $03FF, $07FF, $0FFF,
       $1FFF, $3FFF, $7FFF, $FFFF);
  var
    Size		,    Chunk		: integer;
  begin
    if (OutputBits > 0) then
      OutputBucket := (OutputBucket AND BitBucketMask[OutputBits]) OR
        (longInt(code) SHL OutputBits)
    else
      OutputBucket := code;
    inc(OutputBits, BitsPerCode);
    while (OutputBits >= 8) do
    begin
      WriteByte(Buffer, OutputBucket AND $FF);
      OutputBucket := OutputBucket SHR 8;
      dec(OutputBits, 8);
      if (Buffer.Position >= 254) then
      begin
        Size := Buffer.Position;
        Buffer.Seek(0, soFromBeginning);
        while (Size > 0) do
        begin
          if (Size >= 254) then
            Chunk := 254
          else
            Chunk := Size;
          WriteByte(Stream, Chunk);
          dec(Size, Chunk);
          Stream.CopyFrom(Buffer, Chunk);
        end;
        Buffer.Seek(0, soFromBeginning);
      end;
    end;
    if (FreeEntry > MaxCode) or (ClearFlag) then
    begin
      if (ClearFlag) then
      begin
        BitsPerCode := InitialBitsPerCode;
        MaxCode := MaxCodesFromBits(BitsPerCode);
        ClearFlag := FALSE;
      end else
      begin
        inc(BitsPerCode);
        if (BitsPerCode = GIFCodeBits) then
          MaxCode := TableMaxMaxCode
        else
          MaxCode := MaxCodesFromBits(BitsPerCode);
      end;
    end;
    if (code = EOFCode) then
    begin
      while (OutputBits > 0) do
      begin
        WriteByte(Buffer, OutputBucket AND $FF);
        OutputBucket := OutputBucket SHR 8;
        dec(OutputBits, 8);
      end;
      if (Buffer.Position > 0) then
      begin
        Size := Buffer.Position;
        Buffer.Seek(0, soFromBeginning);
        while (Size > 0) do
        begin
          if (Size >= 254) then
            Chunk := 254
          else
            Chunk := Size;
          WriteByte(Stream, Chunk);
          dec(Size, Chunk);
          Stream.CopyFrom(Buffer, Chunk);
        end;
        Buffer.Seek(0, soFromBeginning);
      end;
    end;
  end;
  procedure HashClear;
  begin
    HashTable.Clear;
    FreeEntry := ClearCode + 2;
    ClearFlag := TRUE;
    output(ClearCode);
  end;
/////////////////////////////////////////
begin
	fdatasize:=width*height;
   bpp := BitsPerPixel;
  if (bpp <= 1) then
  begin
    InitialBitsPerCode := 3;
    WriteByte(Stream, 2);
  end else
  begin
    InitialBitsPerCode := bpp + 1;
    WriteByte(Stream, bpp);
  end;
  ClearFlag := FALSE;
  BitsPerCode := InitialBitsPerCode;
  MaxCode := MaxCodesFromBits(BitsPerCode);
  OutputBucket := 0;
  OutputBits  := 0;
  HashTable := THashTable.Create;
  try
    ClearCode := (1 SHL (InitialBitsPerCode - 1));
    EOFCode := ClearCode + 1;
    Buffer := TMemoryStream.Create;
    try
      Buffer.SetSize(256);
      HashClear;
      if (Interlaced) then
        cX := width
      else
        cX := FDataSize;
      Y := 0;
      Pass := 0;
      if (FData <> nil) and (FDataSize > 0) then
      begin
        //
        Pixel := FData;
        Prefix := CodeInt(Pixel^);
        inc(Pixel);
        while (BumpPixel) do
        begin
          PixelVal := Pixel^;
          inc(Pixel);
          NewKey := (KeyInt(Prefix) SHL 8) OR ord(PixelVal);
          NewCode := HashTable.Lookup(NewKey);
          if (NewCode >= 0) then
          begin
            Prefix := NewCode;
            continue;
          end;
          Output(Prefix);
          Prefix := CodeInt(PixelVal);
          if (FreeEntry < TableMaxFill) then
          begin
            HashTable.Insert(NewKey, FreeEntry);
            inc(FreeEntry);
          end else
            HashClear;
        end;
        output(Prefix);
      end;
      output(EOFCode);
    finally
      Buffer.Free;
    end;
  finally
    HashTable.Free;
  end;
  WriteByte(Stream, EndBlockByte);
end;

(*****************************************************************************************)
(*****************************************************************************************)


end.
